#include "Arduino.h"
#include "IoTool.h"

#if defined(__CPU_ARC__) /* Do not compile this on AVR based boards */
#include "CurieBLESerial.h"
#endif

/**
 * IoToolSensor
 */
IoToolSensor::IoToolSensor(const char id[], const char name[], const char sname[])
{
  _id = id;
  _name = name;
  _sname = sname;
  _units = "";
  _decimals = 0;
}

IoToolSensor::IoToolSensor(const char id[], const char name[], const char sname[], const char units[])
{
  _id = id;
  _name = name;
  _sname = sname;
  _units = units;
  _decimals = 0;
}

IoToolSensor::IoToolSensor(const char id[], const char name[], const char sname[], const char units[], int decimals)
{
  _id = id;
  _name = name;
  _sname = sname;
  _units = units;
  _decimals = decimals;
}

/* Set reference to Stream object, which is used in write method */
void IoToolSensor::setSerial(Stream &serial)
{
  _serial = &serial;
}

void IoToolSensor::getSensorString(char string[])
{
  strcpy(string, SENS_PKT_START);
  strcat(string, _id);
  strcat(string, P_DEM_S);
  strcat(string, _name);
  strcat(string, P_DEM_S);

  /* Take only first 8 characters of user input */
  char s_name_l[9];
  sprintf(s_name_l, "%.8s", _sname);

  strcat(string, s_name_l);
  strcat(string, P_DEM_S);

  /* Take only first 8 characters of user input */
  char s_unit_l[9];
  sprintf(s_unit_l, "%.8s", _units);

  strcat(string, s_unit_l);
  strcat(string, P_DEM_S);
  
  char dec_s[8];
  sprintf(dec_s, "%d", _decimals);
  strcat(string, dec_s);
  strcat(string, P_END_S);
}

/**
 * IoToolAction 
 */
IoToolAction::IoToolAction(const char id[], const char name[])
{
  _id = id;
  _name = name;
}

void IoToolAction::setHandler(void (*callback)(void))
{
  _callback = callback;
}

void IoToolAction::execute(void)
{
  _callback();
}

const char *IoToolAction::getId()
{
  return _id;
}

void IoToolAction::getActionString(char string[])
{
  strcpy(string, ACT_PKT_START);
  strcat(string, _id);
  strcat(string, P_DEM_S);
  strcat(string, _name);
  strcat(string, P_END_S);
}

/**
 * IoTool
 */
IoTool::IoTool()
{
  _aIdx = 0;
  _sIdx = 0;
  _pIdx = 0;
  _parserState = READY;
}

void IoTool::setSerial(Stream &serial)
{
  _serial = &serial;
}

#if defined(__CPU_ARC__)
void IoTool::setBLESerial(CurieBLESerial &serial)
{
  _useBle = true;
  _serial = &serial;
  _curieSerial = (CurieBLESerial *) &serial;
}
#endif

void IoTool::addAction(IoToolAction &action)
{
  if (_aIdx < SENSOR_LIMIT) {
    _actions[_aIdx] = &action;
    _aIdx++;
  }
}

void IoTool::addSensor(IoToolSensor &sensor)
{
  if (_sIdx < SENSOR_LIMIT) {
    sensor.setSerial(*_serial);
    _sensors[_sIdx] = &sensor;
    _sIdx++;
  }
}

void IoTool::pushToBuffer(char c)
{
  _parserBuffer[_pIdx] = c;
  _pIdx++;
}

void IoTool::clearBuffer()
{
  _pIdx = 0;
}

void IoTool::getBufferData(char string[])
{
  strncpy(string, _parserBuffer, _pIdx);
  string[_pIdx] = '\0';
  clearBuffer();
}

/* Used to listen for actions and trigger callbacks */
void IoTool::loop()
{
#if defined(__CPU_ARC__)
  /**
   * 101 BLE specific data retreival process
   */
  if (_useBle) {
    BLECentral central = _curieSerial->getPeripheral()->central();

    if (central && central.connected()) {
      BLECharacteristic *inputCharacteristic = _curieSerial->getInputCharacteristic();

      if (inputCharacteristic->written()) {
	_curieSerial->setBuffer((char *) inputCharacteristic->value());
      }
    }
  }
#endif
  
  while (_serial->available() > 0) {
    char b = _serial->read();

    switch (_parserState) {
    case READY:
      if (b == C_EXEC_ACT) {
	/* Listen for action execute command */
	_parserState = DATA;
      } else if (b == C_SEN_REQ) {
	/* Advertise added sensors */
	_serial->print(C_PACKET_START);

	for (int i = 0; i < _sIdx; i++) {
	  char capability[64];
	  _sensors[i]->getSensorString(capability);
	  _serial->print(capability);
	}
	
	_serial->print(C_PACKET_END);
      } else if (b == C_ACT_REQ) {
	/* Advertise added actions */
	_serial->print(C_PACKET_START);

	for (int i = 0; i < _aIdx; i++) {
	  char action[32];
	  _actions[i]->getActionString(action);
	  _serial->print(action);
	}

	_serial->print(C_PACKET_END);
      }
      break;
    case DATA:
      if (b == P_END) {
	char data[32];
	getBufferData(data);

	/* Find action that needs to be executed and execute it */
	for (int i = 0; i < _aIdx; i++) {
	  const char *id = _actions[i]->getId();

	  if (strcmp(id, data) == 0) {
	    _actions[i]->execute();
	  }
	}
	
	_parserState = READY;	
      } else {
	/* Check if there is enough space left in the buffer */
	if (_pIdx > PARSER_BUFFER_SIZE - 2) {
	  clearBuffer();
	  _parserState = READY;
	  break;
	}
	
	pushToBuffer(b);
      }
      break;
    }
  }
}
