/**
 * IoToolLibrary.h - base library for IoTool Arduino service
 * Created by SenLab, August 24, 2016.
 * TODO: License
 */

#ifndef IOTOOL_H
#define IOTOOL_H

#include "Arduino.h"
#if defined(__CPU_ARC__) /* Do not compile this on AVR based boards */
#include "CurieBLESerial.h"
#endif

#define ACTION_LIMIT 20
#define SENSOR_LIMIT 20

#define PARSER_BUFFER_SIZE 33

#define C_PACKET_START '='
#define C_PACKET_END '^'
#define C_ACT_REQ 'A'
#define C_SEN_REQ 'S'
#define C_EXEC_ACT '#'
#define P_END ';'
#define P_DEM_S ":"
#define P_END_S ";"
#define SENS_PKT_START "$"
#define ACT_PKT_START "&"
#define S_DATA_START "~"

enum ParserState {
  READY, DATA
};

/* Sensor class, multiple sensors can be defined */
class IoToolSensor
{
 public:
  IoToolSensor(const char id[], const char name[], const char sname[]);
  IoToolSensor(const char id[], const char name[], const char sname[], const char units[]);
  IoToolSensor(const char id[], const char name[], const char sname[], const char units[], int decimals);
  template<typename T> void write(T data);
  void setSerial(Stream &serial);
  void getSensorString(char string[]);
 private:
  const char *_id;
  const char *_name;
  const char *_sname;
  const char *_units;
  int _decimals;
  Stream *_serial;
};

template<typename T> void IoToolSensor::write(T data)
{
  /**
   * AVR based arduinos do not support stdio print functions e.g. sprintf 
   * 101 However works as expected. 
   */
#if defined(__CPU_ARC__)
  char temp[48];
  strcpy(temp, S_DATA_START);
  strcat(temp, _id);
  strcat(temp, P_DEM_S);

  /* Get number of decimal places in string format */
  char dec[8];
  sprintf(dec, "%d", _decimals);

  /* Create format string with specified number of decimal places*/
  char fmt[8];
  strcpy(fmt, "%.");
  strcat(fmt, dec);
  strcat(fmt, "f");

  /* Add final value to the data packet */
  char val[32];
  sprintf(val, fmt, (double) data);
  
  strcat(temp, val);
  strcat(temp, P_END_S);

  _serial->print(temp);
#else
  /* This would be inefficient on 101, since updating characteristic takes long time */
  _serial->print(S_DATA_START);
  _serial->print(_id);
  _serial->print(P_DEM_S);
  _serial->print((double) data, _decimals);
  _serial->print(P_END_S);
#endif
}

/* Action class, multiple actions can be defined */
/* Action requires id and corresponding callback function */
class IoToolAction
{
 public:
  IoToolAction(const char id[], const char name[]);
  void setHandler(void (*callback)(void));
  void execute(void);
  const char *getId();
  void getActionString(char string[]);
 private:
  const char *_id;
  const char *_name;
  void (*_callback)(void);
};


/* Main class, only one instance should be active */
/* Listen on provided stream for actions, send attached sensor readings */
class IoTool
{
 public:
  IoTool();
  void setSerial(Stream &serial);
#if defined(__CPU_ARC__)
  void setBLESerial(CurieBLESerial &serial);
#endif
  void addAction(IoToolAction &action);
  void addSensor(IoToolSensor &sensor);
  void loop();
 private:
  boolean _useBle = false;
#if defined(__CPU_ARC__)
  CurieBLESerial* _curieSerial;
#endif
  Stream *_serial;
  IoToolAction *_actions[ACTION_LIMIT];
  int _aIdx;
  IoToolSensor *_sensors[SENSOR_LIMIT];
  int _sIdx;
  void pushToBuffer(char c);
  void clearBuffer();
  void getBufferData(char string[]);
  ParserState _parserState;
  char _parserBuffer[PARSER_BUFFER_SIZE];
  int _pIdx;
};
#endif
